package edu.mit.harold.icollections;

import java.util.AbstractSet;
import java.util.Collection;
import java.util.Iterator;


/**
 * 
 * A map-backed immutable set.
 * 
 * @author harold
 *
 * @param <E>
 */
public final class MapISet<E> extends AbstractSet<E> implements ISet<E> {
//// STATIC FACTORY METHODS ////
	/**
	 * @param <E>
	 * @param map
	 * @return an ISet with the elements of map.keySet(), backed by map
	 */
	@SuppressWarnings("unchecked")
	public static <E> MapISet<E> from(final IMap<E,?> map) {
		return new MapISet<E>((IMap<E,Object>)map, map.keySet().hashCode()); }
	
	/**
	 * @param <E>
	 * @param map
	 * @param e
	 * @return from(map).plus(e)
	 */
	public static <E> MapISet<E> from(final IMap<E,?> map, E e) {
		return from(map).plus(e); }
	
	/**
	 * @param <E>
	 * @param map
	 * @param list
	 * @return from(map).plusAll(list)
	 */
	public static <E> MapISet<E> from(final IMap<E,?> map, final Collection<? extends E> list) {
		return from(map).plusAll(list); }
	

//// PRIVATE CONSTRUCTORS ////
	private final IMap<E,Object> map;
	private final int hashCode;
	// not instantiable (or subclassable):
	private MapISet(final IMap<E,Object> map, final int hashCode) {
		this.map = map; this.hashCode = hashCode; }


//// REQUIRED METHODS FROM AbstractSet ////
	@Override
	public Iterator<E> iterator() {
		return map.keySet().iterator(); }
	@Override
	public int size() {
		return map.size(); }
	

//// OVERRIDDEN METHODS OF AbstractSet ////
	@Override
	public boolean contains(final Object e) {
		return map.containsKey(e); }
	@Override
	public int hashCode() {
		return hashCode; }

	
//// IMPLEMENTED METHODS OF ISet ////
	private static enum In {IN}
	
	public MapISet<E> plus(final E e) {
		if(contains(e)) return this;
		return new MapISet<E>(map.plus(e, In.IN), hashCode+e.hashCode());
	}
	
	public MapISet<E> minus(final Object e) {
		if(!contains(e)) return this;
		return new MapISet<E>(map.minus(e), hashCode-e.hashCode());
	}

	public MapISet<E> plusAll(final Collection<? extends E> list) {
		IMap<E,Object> map = this.map;
		for(E e : list)
			map = map.plus(e, In.IN);
		return from(map); // (completely recomputes hashCode)
	}
	
	public MapISet<E> minusAll(final Collection<?> list) {
		IMap<E,Object> map = this.map.minusAll(list);
		return from(map); // (completely recomputes hashCode)
	}
}
